@@ -1,5 +1,5 @@
 /*
- * Copyright 2011-2014 Samy Al Bahra.
+ * Copyright 2011-2015 Samy Al Bahra.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -24,6 +24,9 @@
  * SUCH DAMAGE.
  */
 
+#ifndef CK_COMMON_H
+#define CK_COMMON_H
+
 #include <ck_cc.h>
 #include <ck_pr.h>
 #include <stdarg.h>
@@ -289,7 +292,7 @@ CK_CC_UNUSED static int
 aff_iterate_core(struct affinity *acb, unsigned int *core)
 {
 	cpu_set_t s;
-	
+
 	*core = ck_pr_faa_uint(&acb->request, acb->delta);
 	CPU_ZERO(&s);
 	CPU_SET((*core) % CORES, &s);
@@ -455,3 +458,9 @@ ck_error(const char *message, ...)
 	exit(EXIT_FAILURE);
 }
 
+#define ck_test(A, B, ...) do {			\
+	if (A)					\
+		ck_error(B, ##__VA_ARGS__);	\
+} while (0)
+
+#endif /* CK_COMMON_H */
