@@ -1,6 +1,6 @@
 /*
- * Copyright 2014 Olivier Houchard
- * Copyright 2012-2014 Samy Al Bahra.
+ * Copyright 2014-2015 Olivier Houchard.
+ * Copyright 2012-2015 Samy Al Bahra.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -31,8 +31,8 @@
 #include <ck_md.h>
 #include <ck_pr.h>
 #include <ck_stdint.h>
-#include <stdbool.h>
-#include <string.h>
+#include <ck_stdbool.h>
+#include <ck_string.h>
 
 #include "ck_internal.h"
 
@@ -78,10 +78,10 @@
 
 enum ck_rhs_probe_behavior {
 	CK_RHS_PROBE = 0,	/* Default behavior. */
-	CK_RHS_PROBE_RH,		/* Short-circuit if RH slot found. */
+	CK_RHS_PROBE_RH,	/* Short-circuit if RH slot found. */
 	CK_RHS_PROBE_INSERT,	/* Short-circuit on probe bound if tombstone found. */
 
-	CK_RHS_PROBE_ROBIN_HOOD,	/* Look for the first slot available for the entry we are about to replace, only used to internally implement Robin Hood */
+	CK_RHS_PROBE_ROBIN_HOOD,/* Look for the first slot available for the entry we are about to replace, only used to internally implement Robin Hood */
 	CK_RHS_PROBE_NO_RH,	/* Don't do the RH dance */
 };
 struct ck_rhs_entry_desc {
@@ -89,7 +89,7 @@ struct ck_rhs_entry_desc {
 	unsigned short wanted;
 	CK_RHS_WORD probe_bound;
 	bool in_rh;
-	void *entry;
+	const void *entry;
 } CK_CC_ALIGN(16);
 
 struct ck_rhs_no_entry_desc {
@@ -105,7 +105,7 @@ typedef long ck_rhs_probe_cb_t(struct ck_rhs *hs,
     long *priority,
     unsigned long h,
     const void *key,
-    void **object,
+    const void **object,
     unsigned long probe_limit,
     enum ck_rhs_probe_behavior behavior);
 
@@ -118,11 +118,12 @@ struct ck_rhs_map {
 	unsigned long n_entries;
 	unsigned long capacity;
 	unsigned long size;
+	unsigned long max_entries;
 	char offset_mask;
 	union {
 		struct ck_rhs_entry_desc *descs;
 		struct ck_rhs_no_entry {
-			void **entries;
+			const void **entries;
 			struct ck_rhs_no_entry_desc *descs;
 		} no_entries;
 	} entries;
@@ -130,7 +131,7 @@ struct ck_rhs_map {
 	ck_rhs_probe_cb_t *probe_func;
 };
 
-static CK_CC_INLINE void *
+static CK_CC_INLINE const void *
 ck_rhs_entry(struct ck_rhs_map *map, long offset)
 {
 
@@ -140,7 +141,7 @@ ck_rhs_entry(struct ck_rhs_map *map, long offset)
 		return (map->entries.descs[offset].entry);
 }
 
-static CK_CC_INLINE void **
+static CK_CC_INLINE const void **
 ck_rhs_entry_addr(struct ck_rhs_map *map, long offset)
 {
 
@@ -242,11 +243,29 @@ ck_rhs_unset_rh(struct ck_rhs_map *map, long offset)
 }
 
 
-#define CK_RHS_LOAD_FACTOR	50
+#define CK_RHS_DEFAULT_LOAD_FACTOR	50
 
 static ck_rhs_probe_cb_t ck_rhs_map_probe;
 static ck_rhs_probe_cb_t ck_rhs_map_probe_rm;
 
+bool
+ck_rhs_set_load_factor(struct ck_rhs *hs, unsigned int load_factor)
+{
+	struct ck_rhs_map *map = hs->map;
+
+	if (load_factor == 0 || load_factor > 100)
+		return false;
+
+	hs->load_factor = load_factor;
+	map->max_entries = (map->capacity * (unsigned long)hs->load_factor) / 100;
+	while (map->n_entries > map->max_entries) {
+		if (ck_rhs_grow(hs, map->capacity << 1) == false)
+			return false;
+		map = hs->map;
+	}
+	return true;
+}
+
 void
 ck_rhs_iterator_init(struct ck_rhs_iterator *iterator)
 {
@@ -266,7 +285,7 @@ ck_rhs_next(struct ck_rhs *hs, struct ck_rhs_iterator *i, void **key)
 		return false;
 
 	do {
-		value = ck_rhs_entry(map, i->offset);
+		value = CK_CC_DECONST_PTR(ck_rhs_entry(map, i->offset));
 		if (value != CK_RHS_EMPTY) {
 #ifdef CK_RHS_PP
 			if (hs->mode & CK_RHS_MODE_OBJECT)
@@ -322,7 +341,7 @@ ck_rhs_map_create(struct ck_rhs *hs, unsigned long entries)
 
 	n_entries = ck_internal_power_2(entries);
 	if (n_entries < CK_RHS_PROBE_L1)
-		return NULL;
+		n_entries = CK_RHS_PROBE_L1;
 
 	if (hs->mode & CK_RHS_MODE_READ_MOSTLY)
 		size = sizeof(struct ck_rhs_map) +
@@ -351,6 +370,7 @@ ck_rhs_map_create(struct ck_rhs *hs, unsigned long entries)
 	map->mask = n_entries - 1;
 	map->n_entries = 0;
 
+	map->max_entries = (map->capacity * (unsigned long)hs->load_factor) / 100;
 	/* Align map allocation to cache line. */
 	if (map->read_mostly) {
 		map->entries.no_entries.entries = (void *)(((uintptr_t)&map[1] +
@@ -359,7 +379,7 @@ ck_rhs_map_create(struct ck_rhs *hs, unsigned long entries)
 		memset(map->entries.no_entries.entries, 0,
 		    sizeof(void *) * n_entries);
 		memset(map->entries.no_entries.descs, 0,
-		    sizeof(struct ck_rhs_no_entry_desc));
+		    sizeof(struct ck_rhs_no_entry_desc) * n_entries);
 		map->offset_mask = (CK_MD_CACHELINE / sizeof(void *)) - 1;
 		map->probe_func = ck_rhs_map_probe_rm;
 
@@ -475,7 +495,7 @@ ck_rhs_grow(struct ck_rhs *hs,
     unsigned long capacity)
 {
 	struct ck_rhs_map *map, *update;
-	void *previous, *prev_saved;
+	const void *previous, *prev_saved;
 	unsigned long k, offset, probes;
 
 restart:
@@ -504,7 +524,7 @@ restart:
 		probes = 0;
 
 		for (;;) {
-			void **cursor = ck_rhs_entry_addr(update, offset);
+			const void **cursor = ck_rhs_entry_addr(update, offset);
 
 			if (probes++ == update->probe_limit) {
 				/*
@@ -522,7 +542,7 @@ restart:
 				ck_rhs_map_bound_set(update, h, probes);
 				break;
 			} else if (ck_rhs_probes(update, offset) < probes) {
-				void *tmp = prev_saved;
+				const void *tmp = prev_saved;
 				unsigned int old_probes;
 				prev_saved = previous = *cursor;
 #ifdef CK_RHS_PP
@@ -563,11 +583,11 @@ ck_rhs_map_probe_rm(struct ck_rhs *hs,
     long *priority,
     unsigned long h,
     const void *key,
-    void **object,
+    const void **object,
     unsigned long probe_limit,
     enum ck_rhs_probe_behavior behavior)
 {
-	void *k;
+	const void *k;
 	const void *compare;
 	long pr = -1;
 	unsigned long offset, probes, opl;
@@ -675,11 +695,11 @@ ck_rhs_map_probe(struct ck_rhs *hs,
     long *priority,
     unsigned long h,
     const void *key,
-    void **object,
+    const void **object,
     unsigned long probe_limit,
     enum ck_rhs_probe_behavior behavior)
 {
-	void *k;
+	const void *k;
 	const void *compare;
 	long pr = -1;
 	unsigned long offset, probes, opl;
@@ -782,24 +802,25 @@ leave:
 	return offset;
 }
 
-static inline void *
+static inline const void *
 ck_rhs_marshal(unsigned int mode, const void *key, unsigned long h)
 {
-	void *insert;
-
 #ifdef CK_RHS_PP
+	const void *insert;
+
 	if (mode & CK_RHS_MODE_OBJECT) {
 		insert = (void *)((uintptr_t)CK_RHS_VMA(key) | ((h >> 25) << CK_MD_VMA_BITS));
 	} else {
-		insert = (void *)key;
+		insert = key;
 	}
+
+	return insert;
 #else
 	(void)mode;
 	(void)h;
-	insert = (void *)key;
-#endif
 
-	return insert;
+	return key;
+#endif
 }
 
 bool
@@ -886,20 +907,21 @@ ck_rhs_put_robin_hood(struct ck_rhs *hs,
     long orig_slot, struct ck_rhs_entry_desc *desc)
 {
 	long slot, first;
-	void *object, *insert;
+	const void *object, *insert;
 	unsigned long n_probes;
 	struct ck_rhs_map *map;
 	unsigned long h = 0;
 	long prev;
 	void *key;
-	long prevs[512];
+	long prevs[CK_RHS_MAX_RH];
 	unsigned int prevs_nb = 0;
+	unsigned int i;
 
 	map = hs->map;
 	first = orig_slot;
 	n_probes = desc->probes;
 restart:
-	key = ck_rhs_entry(map, first);
+	key = CK_CC_DECONST_PTR(ck_rhs_entry(map, first));
 	insert = key;
 #ifdef CK_RHS_PP
 	if (hs->mode & CK_RHS_MODE_OBJECT)
@@ -915,10 +937,13 @@ restart:
 	if (slot == -1 && first == -1) {
 		if (ck_rhs_grow(hs, map->capacity << 1) == false) {
 			desc->in_rh = false;
-			for (unsigned int i = 0; i < prevs_nb; i++)
+
+			for (i = 0; i < prevs_nb; i++)
 				ck_rhs_unset_rh(map, prevs[i]);
+
 			return -1;
 		}
+
 		return 1;
 	}
 
@@ -1047,7 +1072,8 @@ ck_rhs_fas(struct ck_rhs *hs,
     void **previous)
 {
 	long slot, first;
-	void *object, *insert;
+	const void *object;
+	const void *insert;
 	unsigned long n_probes;
 	struct ck_rhs_map *map = hs->map;
 	struct ck_rhs_entry_desc *desc, *desc2;
@@ -1085,7 +1111,123 @@ restart:
 		ck_pr_store_ptr(ck_rhs_entry_addr(map, slot), insert);
 		ck_rhs_set_probes(map, slot, n_probes);
 	}
-	*previous = object;
+	*previous = CK_CC_DECONST_PTR(object);
+	return true;
+}
+
+/*
+ * An apply function takes two arguments. The first argument is a pointer to a
+ * pre-existing object. The second argument is a pointer to the fifth argument
+ * passed to ck_hs_apply. If a non-NULL pointer is passed to the first argument
+ * and the return value of the apply function is NULL, then the pre-existing
+ * value is deleted. If the return pointer is the same as the one passed to the
+ * apply function then no changes are made to the hash table.  If the first
+ * argument is non-NULL and the return pointer is different than that passed to
+ * the apply function, then the pre-existing value is replaced. For
+ * replacement, it is required that the value itself is identical to the
+ * previous value.
+ */
+bool
+ck_rhs_apply(struct ck_rhs *hs,
+    unsigned long h,
+    const void *key,
+    ck_rhs_apply_fn_t *fn,
+    void *cl)
+{
+	const void *insert;
+	const void  *object, *delta = false;
+	unsigned long n_probes;
+	long slot, first;
+	struct ck_rhs_map *map;
+	bool delta_set = false;
+
+restart:
+	map = hs->map;
+
+	slot = map->probe_func(hs, map, &n_probes, &first, h, key, &object, map->probe_limit, CK_RHS_PROBE_INSERT);
+	if (slot == -1 && first == -1) {
+		if (ck_rhs_grow(hs, map->capacity << 1) == false)
+			return false;
+
+		goto restart;
+	}
+	if (!delta_set) {
+		delta = fn(CK_CC_DECONST_PTR(object), cl);
+		delta_set = true;
+	}
+
+	if (delta == NULL) {
+		/*
+		 * The apply function has requested deletion. If the object doesn't exist,
+		 * then exit early.
+		 */
+		if (CK_CC_UNLIKELY(object == NULL))
+			return true;
+
+		/* Otherwise, delete it. */
+		ck_rhs_do_backward_shift_delete(hs, slot);
+		return true;
+	}
+
+	/* The apply function has not requested hash set modification so exit early. */
+	if (delta == object)
+		return true;
+
+	/* A modification or insertion has been requested. */
+	ck_rhs_map_bound_set(map, h, n_probes);
+
+	insert = ck_rhs_marshal(hs->mode, delta, h);
+	if (first != -1) {
+		/*
+		 * This follows the same semantics as ck_hs_set, please refer to that
+		 * function for documentation.
+		 */
+		struct ck_rhs_entry_desc *desc = NULL, *desc2;
+		if (slot != -1) {
+			desc = ck_rhs_desc(map, slot);
+			desc->in_rh = true;
+		}
+		desc2 = ck_rhs_desc(map, first);
+		int ret = ck_rhs_put_robin_hood(hs, first, desc2);
+		if (slot != -1)
+			desc->in_rh = false;
+
+		if (CK_CC_UNLIKELY(ret == 1))
+			goto restart;
+		if (CK_CC_UNLIKELY(ret == -1))
+			return false;
+		/* If an earlier bucket was found, then store entry there. */
+		ck_pr_store_ptr(ck_rhs_entry_addr(map, first), insert);
+		desc2->probes = n_probes;
+		/*
+		 * If a duplicate key was found, then delete it after
+		 * signaling concurrent probes to restart. Optionally,
+		 * it is possible to install tombstone after grace
+		 * period if we can guarantee earlier position of
+		 * duplicate key.
+		 */
+		ck_rhs_add_wanted(hs, first, -1, h);
+		if (object != NULL) {
+			ck_pr_inc_uint(&map->generation[h & CK_RHS_G_MASK]);
+			ck_pr_fence_atomic_store();
+			ck_rhs_do_backward_shift_delete(hs, slot);
+		}
+	} else {
+		/*
+		 * If we are storing into same slot, then atomic store is sufficient
+		 * for replacement.
+		 */
+		ck_pr_store_ptr(ck_rhs_entry_addr(map, slot), insert);
+		ck_rhs_set_probes(map, slot, n_probes);
+		if (object == NULL)
+			ck_rhs_add_wanted(hs, slot, -1, h);
+	}
+
+	if (object == NULL) {
+		map->n_entries++;
+		if ((map->n_entries ) > map->max_entries)
+			ck_rhs_grow(hs, map->capacity << 1);
+	}
 	return true;
 }
 
@@ -1096,7 +1238,8 @@ ck_rhs_set(struct ck_rhs *hs,
     void **previous)
 {
 	long slot, first;
-	void *object, *insert;
+	const void *object;
+	const void *insert;
 	unsigned long n_probes;
 	struct ck_rhs_map *map;
 
@@ -1160,11 +1303,11 @@ restart:
 
 	if (object == NULL) {
 		map->n_entries++;
-		if ((map->n_entries ) > ((map->capacity * CK_RHS_LOAD_FACTOR) / 100))
+		if ((map->n_entries ) > map->max_entries)
 			ck_rhs_grow(hs, map->capacity << 1);
 	}
 
-	*previous = object;
+	*previous = CK_CC_DECONST_PTR(object);
 	return true;
 }
 
@@ -1175,7 +1318,8 @@ ck_rhs_put_internal(struct ck_rhs *hs,
     enum ck_rhs_probe_behavior behavior)
 {
 	long slot, first;
-	void *object, *insert;
+	const void *object;
+	const void *insert;
 	unsigned long n_probes;
 	struct ck_rhs_map *map;
 
@@ -1218,7 +1362,7 @@ restart:
 	}
 
 	map->n_entries++;
-	if ((map->n_entries ) > ((map->capacity * CK_RHS_LOAD_FACTOR) / 100))
+	if ((map->n_entries ) > map->max_entries)
 		ck_rhs_grow(hs, map->capacity << 1);
 	return true;
 }
@@ -1247,7 +1391,7 @@ ck_rhs_get(struct ck_rhs *hs,
     const void *key)
 {
 	long first;
-	void *object;
+	const void *object;
 	struct ck_rhs_map *map;
 	unsigned long n_probes;
 	unsigned int g, g_p, probe;
@@ -1267,7 +1411,7 @@ ck_rhs_get(struct ck_rhs *hs,
 		g_p = ck_pr_load_uint(generation);
 	} while (g != g_p);
 
-	return object;
+	return CK_CC_DECONST_PTR(object);
 }
 
 void *
@@ -1276,7 +1420,7 @@ ck_rhs_remove(struct ck_rhs *hs,
     const void *key)
 {
 	long slot, first;
-	void *object;
+	const void *object;
 	struct ck_rhs_map *map = hs->map;
 	unsigned long n_probes;
 
@@ -1287,7 +1431,7 @@ ck_rhs_remove(struct ck_rhs *hs,
 
 	map->n_entries--;
 	ck_rhs_do_backward_shift_delete(hs, slot);
-	return object;
+	return CK_CC_DECONST_PTR(object);
 }
 
 bool
@@ -1304,6 +1448,7 @@ ck_rhs_move(struct ck_rhs *hs,
 	hs->mode = source->mode;
 	hs->seed = source->seed;
 	hs->map = source->map;
+	hs->load_factor = source->load_factor;
 	hs->m = m;
 	hs->hf = hf;
 	hs->compare = compare;
@@ -1328,6 +1473,7 @@ ck_rhs_init(struct ck_rhs *hs,
 	hs->seed = seed;
 	hs->hf = hf;
 	hs->compare = compare;
+	hs->load_factor = CK_RHS_DEFAULT_LOAD_FACTOR;
 
 	hs->map = ck_rhs_map_create(hs, n_entries);
 	return hs->map != NULL;
